'	**********************************************************************************************
'	*                                                                                            *
'	*	      PNut Booter                                                                    *
'	*                                                                                            *
'	*	      Version 0.1   12/10/2004                                                       *
'	*                                                                                            *
'	*	      Copyright 2004, 2014 Parallax Inc.                                             *
'	*                                                                                            *
'	*	      This file is part of the hardware description for the Propeller 1 Design.      *
'	*                                                                                            *
'	*	      The Propeller 1 Design is free software: you can redistribute it and/or modify *
'	*	      it under the terms of the GNU General Public License as published by the       *
'	*	      Free Software Foundation, either version 3 of the License, or (at your option) *
'	*	      any later version.                                                             *
'	*                                                                                            *
'	*	      The Propeller 1 Design is distributed in the hope that it will be useful,      *
'	*	      but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
'	*	      or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
'	*	      more details.                                                                  *
'	*                                                                                            *
'	*	      You should have received a copy of the GNU General Public License along with   *
'	*	      the Propeller 1 Design.  If not, see <http://www.gnu.org/licenses/>.           *
'	*                                                                                            *
'	**********************************************************************************************
'
'
' Entry
'
PUB main

CON

  #5, masklong, masktop, maskwr				'constants
  lsb, id, dcall, pbase, vbase, dbase, pcurr, dcurr	'variables

  stack = masklong	'$FFFFFFFF			'aliases
  subid = masktop	'$80000000
  parms = maskwr	'$00800000
  psave = lsb

  regop = $3F, r = $80, w = $A0, o = $C0

  interp = $F004

  SPI_MISO_PIN = 8      ' SPI FLASH pins on port A
  SPI_CLK_PIN  = 9
  SPI_MOSI_PIN = 10
  SPI_CS_PIN   = 11

  BOOTCLK       = 12_500_000 ' 12.5MHz
'
'
' Start code
'

DAT			org

entry
			test	mask_rx,ina	wc	'if rx high, check for host
	if_nc		jmp	#boot			'else, boot from eeprom

			call	#rx_bit			'measure rx calibration pulses ($F9)
			mov	threshold,delta		'and calculate threshold
			call	#rx_bit			'(any timeout results in eeprom boot)
			add	threshold,delta
			shr	threshold,#1

			mov	count,#250		'ready to receive/verify 250 lfsr bits
:lfsrin			call	#rx_bit			'receive bit ($FE/$FF) into c
			muxc	lfsr,#$100		'compare to lfsr lsb
			test	lfsr,#$101	wc
	if_c		jmp	#boot			'if mismatch, boot from eeprom
			test	lfsr,#$B2	wc	'advance lfsr
			rcl	lfsr,#1
			djnz	count,#:lfsrin

			or	outa,mask_tx		'host present, make tx high output
			or	dira,mask_tx

			mov	count,#250		'ready to transmit 250 lfsr bits
:lfsrout		test	lfsr,#$01	wz	'send lfsr bit ($FE/$FF)
			call	#tx_bit
			test	lfsr,#$B2	wc	'advance lfsr
			rcl	lfsr,#1
			djnz	count,#:lfsrout

			rdbyte	bits,hFFFF		'get version byte at $FFFF
			mov	count,#8		'send version byte
:version		test	bits,#$01	wz
			call	#tx_bit
			shr	bits,#1
			djnz	count,#:version

			call	#rx_long		'receive command
			mov	command,rxdata

			tjz	command,#shutdown	'if command 0, shutdown
			cmp	command,#4	wc	'if command 4+, shutdown
	if_nc		jmp	#shutdown

			call	#rx_long		'get long count
			mov	count,rxdata

			mov	address,#0		'get longs into ram
:longs			call	#rx_long
			wrlong	rxdata,address
			add	address,#4
			djnz	count,#:longs

			mov	count,h8000		'zero remainder of ram
			sub	count,address
			shr	count,#2	wz
:zero	if_nz		wrlong	zero,address
	if_nz		add	address,#4
	if_nz		djnz	count,#:zero		'(count=0, address=$8000)

			rdword	bits,#$0004+6		'get dbase address
			sub	bits,#4			'set pcurr to $FFF9
			wrlong	hFFF9FFFF,bits
			sub	bits,#4			'set pbase flags
			wrlong	hFFF9FFFF,bits

			mov	bits,#0			'compute ram checksum
:checksum		rdbyte	rxdata,count
			add	bits,rxdata
			add	count,#1
			djnz	address,#:checksum
			test	bits,#$FF	wz	'z=1 if checksum okay

			call	#tx_bit_align		'send checksum okay/error

	if_nz		jmp	#shutdown		'if checksum error, shutdown

			djnz	command,#program	'if command 2-3, program eeprom
			jmp	#launch			'else command 1, launch
'
'
' Program and verify flash from ram
'
program			mov	smode,#1		'set mode in case error
                        call    #spi_init
			mov	address,#0		'reset address for erasing and writing pages
                        call    #spi_erase              'erase 32kB block
        if_z            jmp     #shutdown               'check for erase timeout error
:page			call	#spi_write		'send program command
			mov	count,#$100		'page-program $100 bytes
:byte			rdbyte	eedata,address		'get ram byte
			call	#spi_transfer		'send ram byte
			add	address,#1		'inc address
			djnz	count,#:byte		'loop until page sent
			call	#spi_stop		'initiate page-program cycle
                        call    #spi_check_timeout      'check write timeout status
	if_z		jmp	#shutdown		'if write error, shutdown
			cmp	address,h8000	wz	'check for address $8000
	if_nz		jmp	#:page			'loop until done (z=1 after)

			call	#tx_bit_align		'program done, send okay (z=1)

			call	#spi_read		'send read command
                        mov     count, h8000
:verify			call	#spi_transfer		'get flash byte
			rdbyte	bits,address		'get ram byte
			cmp	bits,eedata	wz	'compare bytes
	if_nz		jmp	#shutdown		'if verify error, shutdown (sends error)
			add	address,#1		'inc address
			djnz	count,#:verify		'loop until done
			call	#spi_stop		'end read (z=1 from before)

			call	#tx_bit_align		'verify done, send okay (z=1)

			mov	smode,#0		'clear mode in case error

			djnz	command,#launch		'if command 3, launch
			jmp	#shutdown		'else command 2, shutdown
'
'
' Load ram from flash and launch
'
boot			mov	smode,#0		'clear mode in case error
                        call    #spi_init
			call	#spi_read		'send read command
                        mov     count, h8000
:loop			call	#spi_transfer		'get spi byte
			wrbyte	eedata,address		'write to ram
			add	address,#1		'inc address
			djnz	count,#:loop		'loop until done
			call	#spi_stop		'end read (followed by launch)
'
'
' Launch program in ram
'
launch			rdword	address,#$0004+2	'if pbase address invalid, shutdown
			cmp	address,#$0010	wz
	if_nz		jmp	#shutdown

			rdbyte	address,#$0004		'if xtal/pll enabled, start up now
			and	address,#$F8		'..while remaining in rcfast mode
			clkset	address

:delay			djnz	time_xtal,#:delay	'allow 20ms @20MHz for xtal/pll to settle

			rdbyte	address,#$0004		'switch to selected clock
			clkset	address

			coginit	interpreter		'reboot cog with interpreter
'
'
' Shutdown
'
shutdown
:call			call	#spi_stop		'ensure spi transfer is complete

			cmp	smode,#0	wz	'if serial mode, send error (z=0)
	if_nz		mov	smode,#0		'(only do once)
	if_nz		mov	:call,#0		'(replace call with nop)
	if_nz		call	#tx_bit_align		'(may return to shutdown, no problem)

			mov	dira,#0			'cancel any outputs

			mov	smode,#$02		'stop clock
			clkset	smode			'(suspend until reset)
hang                    jmp     #hang

'***************************************
'* SPI routines for W25Q80DV 1MB flash *
'***************************************

spi_init                or      outa, spimask
                        or      dira, spimask           ' begin with all SPI output pins high
spi_init_ret            ret

'***************************************

spi_transfer            mov     bits, #8 wc             ' 8 bit tranfers, clear carry
                        shl     eedata, #24             ' move data into top byte of long
                        rcl     eedata, #1 wc           ' start with MSB of output in C
spi_loop                muxc    outa, spimaskmosi       ' output data bit to MOSI pin
                        andn    outa, spimaskclk        ' drop clk low
                        or      outa, spimaskclk        ' raise clk high
                        test    spimaskmiso, ina wc, nr ' read input data bit from MISO pin
                        rcl     eedata, #1 wc           ' advance and store data
                        djnz    bits, #spi_loop         ' repeat 8 times
                        or      outa, spimaskmosi       ' let MOSI pin idle high
spi_transfer_ret        ret

'***************************************

spi_stop                or      outa, spimask           ' bring chip select (and other pins) high
spi_stop_ret            ret

'***************************************

spi_read                mov     address,#0              ' reset read address for reading or verifying
                        mov     eedata, #3              ' send flash READ command
spi_send_cmd_addr       andn    outa, spimaskcs         ' drop chip select low
                        call    #spi_transfer
                        mov     eedata, #$0F            ' address24:address16 set to upper block
                        call    #spi_transfer
                        mov     eedata, address
                        shr     eedata, #8              ' get flash page to read or write
                        add     eedata, #$80            ' address15:address8 (add page to flash offset $f8000)
                        call    #spi_transfer
                        mov     eedata, #00             ' address7:address0
                        call    #spi_transfer
spi_send_cmd_addr_ret
spi_read_ret            ret

'***************************************

spi_write_enable        andn    outa, spimaskcs         ' drop cs low
                        mov     eedata, #6              ' send flash WRITE ENABLE command
                        call    #spi_transfer           ' to enable flash access
                        call    #spi_stop
spi_write_enable_ret    ret

'***************************************

spi_write               call    #spi_write_enable       ' enable writes before writing
                        mov     eedata,#2               ' send flash WRITE command
                        call    #spi_send_cmd_addr      ' also send current address to write
spi_write_ret           ret

'***************************************

spi_erase               call    #spi_write_enable       ' enable writes before erase
                        mov     eedata, #$52            ' send flash ERASE command
                        call    #spi_send_cmd_addr      ' erase upper 32kB block in flash
                        call    #spi_stop               ' finish SPI access
spi_check_timeout       mov     eedata, #5              ' read status register
                        andn    outa, spimaskcs         ' drop chip select low
                        call    #spi_transfer           ' polling loop takes 58 instructions x 4 clocks/ins / 12.5MHz = 18.56us per iteration
                        mov     timer, erasetimeout     ' Set timeout to ~850ms. 32kB erase should take 800ms max, page write is faster
spi_poll                call    #spi_transfer
                        test    eedata, #1 wz           ' z=1 if done
if_nz                   djnz    timer, #spi_poll
                        or      timer, timer wz         ' z=1 if timeout
                        call    #spi_stop
spi_check_timeout_ret
spi_erase_ret           ret

'***************************************


'
'
'*******************
'* Serial routines *
'*******************
'
'
' Transmit bit (nz)
' conveys incoming $F9 on rx to $FE/$FF on tx
'
tx_bit_align		mov	time,time_load		'reset time limit

:align			call	#rx_bit			'align to next $F9
	if_c		jmp	#:align

tx_bit			mov	time,time_load		'reset time limit

:high			test	mask_rx,ina	wc	'wait while high
	if_c		djnz	time,#:high
	if_c		jmp	#shutdown		'if timeout, shutdown

			andn	outa,mask_tx		'tx low

:low			test	mask_rx,ina	wc	'wait while low
	if_nc		djnz	time,#:low
	if_nc		jmp	#shutdown		'if timeout, shutdown

			muxnz	outa,mask_tx		'tx low/high

:high2			test	mask_rx,ina	wc	'wait while high
	if_c		djnz	time,#:high2
	if_c		jmp	#shutdown		'if timeout, shutdown

			or	outa,mask_tx		'tx high

:low2			test	mask_rx,ina	wc	'wait while low
	if_nc		djnz	time,#:low2
	if_nc		jmp	#shutdown		'if timeout, shutdown

tx_bit_ret
tx_bit_align_ret	ret
'
'
' Receive long
'
rx_long			mov	time,time_load		'reset time limit

			mov	bits,#32		'ready for 32 bits
:loop			call	#rx_bit			'input bit
			rcr	rxdata,#1		'shift into long
			djnz	bits,#:loop		'loop until done

rx_long_ret		ret
'
'
' Receive bit (c)
'
rx_bit			test	mask_rx,ina	wc	'wait while rx high
	if_c		djnz	time,#rx_bit
	if_c		jmp	#boot			'if timeout, boot from eeprom

			mov	delta,time		'time while rx low

:loop			test	mask_rx,ina	wc	'h?h	2 instructions/loop
	if_nc		djnz	time,#:loop		'1	400ns @20MHz...1us @8MHz
	if_nc		jmp	#boot			'if timeout, boot from eeprom

			sub	delta,time		'delta = rx low time in loops
			cmp	delta,threshold	wc	'resolve bit into c

rx_bit_ret		ret
'
'
' Constants
'
mask_rx			long	$80000000
mask_tx			long	$40000000
spimask                 long    (1<<SPI_MOSI_PIN) + (1<<SPI_CLK_PIN) + (1<<SPI_CS_PIN)
spimaskcs               long    (1<<SPI_CS_PIN)
spimaskmosi             long    (1<<SPI_MOSI_PIN)
spimaskmiso             long    (1<<SPI_MISO_PIN)
spimaskclk              long    (1<<SPI_CLK_PIN)
hFFFF                   long    $ffff
erasetimeout            long    (((BOOTCLK / 1000) * 850) / (4*58)) ' 45797 iterations @ 12.5MHz
time			long	150 * 20000 / 4 / 2	'150ms (@20MHz, 2 inst/loop)
time_load		long	100 * 20000 / 4 / 2	'100ms (@20MHz, 2 inst/loop)
time_xtal		long	20 * 20000 / 4 / 1	'20ms (@20MHz, 1 inst/loop)
lfsr			long	"P"
zero			long	0
smode			long	0
hFFF9FFFF		long	$FFF9FFFF
h8000			long	$8000
interpreter		long	$0001 << 18 + $3C01 << 4 + %0000

'
'
' Variables
'
vars_start              ' keep this symbol at start of reserved variable space

command			res	1
address			res	1
count			res	1
bits			res	1
eedata			res	1
rxdata			res	1
delta			res	1
threshold		res	1
timer                   res     1

                        org     vars_start
                        long    0[$1C0 - $]
' Copyright
                        byte    "================"
                        byte    " Copyright 2005 "
                        byte    " Parallax, Inc. "
                        byte    "----------------"
                        byte    "www.parallax.com"
                        byte    "================"
' SPIN boot code
                        long    0[$200 - $24 - $]

'	**********************************************************************************************
'	*                                                                                            *
'	*       PNut Runner                                                                          *
'	*                                                                                            *
'	*       Version 0.1   12/10/2004                                                             *
'	*                                                                                            *
'	*	      Copyright 2004, 2014 Parallax Inc.                                             *
'	*                                                                                            *
'	*	      This file is part of the hardware description for the Propeller 1 Design.      *
'	*                                                                                            *
'	*	      The Propeller 1 Design is free software: you can redistribute it and/or modify *
'	*	      it under the terms of the GNU General Public License as published by the       *
'	*	      Free Software Foundation, either version 3 of the License, or (at your option) *
'	*	      any later version.                                                             *
'	*                                                                                            *
'	*	      The Propeller 1 Design is distributed in the hope that it will be useful,      *
'	*	      but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
'	*	      or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   *
'	*	      more details.                                                                  *
'	*                                                                                            *
'	*	      You should have received a copy of the GNU General Public License along with   *
'	*	      the Propeller 1 Design.  If not, see <http://www.gnu.org/licenses/>.           *
'	*                                                                                            *
'	**********************************************************************************************
'
'
' Run code
' - builds new stack and launches interpreter
' - jumped to by 'run' instruction
'

first	byte	regop, w + stack		'pop and write stack (entry)
	byte	regop, o + subid, $80		'pop and write subid (leave on stack)

	byte	$38, 8, $E2, $38, 2, $E3	'parms := subid >> 8 << 2
	byte	regop, w + parms

	byte	regop, r + subid		'subid := @pbase.long[subid & $FF]
	byte	$38, $FF, $E8, $D7, $00
	byte	regop, w + subid

	byte	regop, r + stack, $38, 12, $EC	'bytemove(stack + 12, dcurr - 4 - parms, parms)
	byte	regop, r + dcurr
	byte	$38, 4, $ED
	byte	regop, r + parms, $ED
	byte	regop, r + parms, $1C

	byte	regop, r + parms, $14		'pop(parms)

	byte	$34				'long[stack] := $FFFFFFFF
	byte	regop, r + stack, $C1

	byte	$3B, $FF, $F9, $FF, $FF		'long[stack][1] := $FFF9FFFF
	byte	regop, r + stack, $36, $D1

	byte	$35				'long[stack][2] := 0
	byte	regop, r + stack, $38, 2, $D1

	byte	regop, r + stack, $38, 12, $EC	'parms += stack + 12
	byte	regop, o + parms, $4C

	byte	regop, r + pbase		'word[parms][1] := pbase
	byte	regop, r + parms, $36, $B1

	byte	regop, r + vbase		'word[parms][2] := vbase
	byte	regop, r + parms, $38, 2, $B1

	byte	regop, r + stack, $38, 8, $EC	'word[parms][3] := stack + 8
	byte	regop, r + parms, $38, 3, $B1

	byte	regop, r + subid, $A0, $97, $00	'word[parms][4] := @pbase.byte[word[subid]]
	byte	regop, r + parms, $38, 4, $B1

	byte	regop, r + subid, $36, $B0	'word[parms][5] := parms + word[subid][1]
	byte	regop, r + parms, $EC
	byte	regop, r + parms, $38, 5, $B1

	byte	$34				'push params for coginit(-1, interp, parms)
	byte	$39, interp >> 8, interp
	byte	regop, r + parms

	byte	regop, r + psave		'push psave

	byte	$34				'restore masklong to $FFFFFFFF
	byte	regop, w + masklong

	byte	$37, $1E			'restore masktop to $80000000
	byte	regop, w + masktop

	byte	$37, $16			'restore maskwr to $00800000
	byte	regop, w + maskwr

	byte	regop, w + pcurr		'pop and write pcurr (return to caller)
'
'
' Position stop code and entry code
'
	long
	byte	0
'
'
' Stop code at $FFF9
'
	byte	regop, r + id, $21		'cogstop(cogid)
'
'
' Entry code at $FFFC
'
	byte	$04				'jmp start
	byte	(@first - @last) >> 8
	byte	@first - @last

last	byte	$02				'chip version


